/* $Id: http.c,v 1.4 2005/08/03 17:07:54 bedlam Exp $ */

#include <signal.h>
#include <curl/curl.h>
#include "in_vorbis.h"
#include "mp3.h"
#ifdef HAVE_LIBFLAC
#include "in_flac.h"
#include <FLAC/stream_decoder.h>
#endif
#include "definitions.h"
#include "metadata.h"


typedef struct http_private_t {
  CURL *curl_handle;
  CURLM *multi_handle;

  struct curl_slist *header_list;
  struct curl_slist *httpaliases_list;
  int icy_metaint;
  int bytes_read;
  char error[CURL_ERROR_SIZE];
  input_stream_t *input;

  int control_metadata; /* prevent from setting too early ogg metadata */

  char *buffer;		/* buffer to store cached data*/
  int buffer_len;	/* currently allocated buffers length */
  int buffer_pos;	/* end of data in buffer*/
  int still_running;	/* Is background url fetch still in progress */

} http_private_t;


extern stream404_config_t stream404_config;

static http_private_t private;

#define METADATAON 1
#define METADATAOFF 0

size_t vorbisfile_cb_read (void *, size_t, size_t, void *);
int vorbisfile_cb_seek (void *, ogg_int64_t, int);
int vorbisfile_cb_close (void *);
long vorbisfile_cb_tell (void *);

int http_open (input_stream_t *, int);
int http_peek (input_stream_t *, void *, size_t, size_t);
int http_read (input_stream_t *, void *, size_t, size_t);
int http_seek (input_stream_t *, long, int);
long http_tell (input_stream_t *);
void http_close (input_stream_t *);



ov_callbacks vorbisfile_callbacks = {
  &vorbisfile_cb_read,
  &vorbisfile_cb_seek,
  &vorbisfile_cb_close,
  &vorbisfile_cb_tell
};

/* ------------------- Vorbisfile Callbacks ----------------- */

size_t vorbisfile_cb_read (void *ptr, size_t size, size_t nmemb, void *arg)
{
  return http_read(arg, ptr, size, nmemb);
}

int vorbisfile_cb_seek (void *arg, ogg_int64_t offset, int whence)
{
  return http_seek(arg, offset, whence);
}
int vorbisfile_cb_close (void *arg)
{
  return 1; /* Ignore close request so transport can be closed later */

  /* XXX soon to be replaced */
  /* http_close(decoder->source) */
}

long vorbisfile_cb_tell (void *arg)
{
  return http_tell(arg);
}


/* --------------- VORBIS PART --------------- */
/* -- static prototypes -- */

extern int stream404_vorbis_readpcm (input_stream_t* self, size_t len,
                                int16_t* left, int16_t* right);
extern int stream404_vorbis_close (input_stream_t* self);
extern void in_vorbis_parse (input_stream_t* self);

int http_open_vorbis(input_stream_t *);
int http_vorbis_close (input_stream_t*);

typedef struct {
  OggVorbis_File* vf;
  vorbis_info* info;
  int link;
  int16_t buf[2048];
  size_t samples;
  int offset;
} stream404_vorbis_in_t;

int http_open_vorbis(input_stream_t *source)
{
  stream404_vorbis_in_t* vorbis_data;
  OggVorbis_File* vf;
  int rc;

  /* this is copied from stream404_open_vorbis XXX should be merged */
  if (http_open(source, METADATAOFF) < 0) {
    stream404_log("Cannot open remote URL");
    return -1;
  }
  
  if (! (vf = (OggVorbis_File*) malloc (sizeof (OggVorbis_File)))) {
    stream404_log_error ("Malloc failed in stream404_vorbis_open");
    return -1;
  }
  
  if ((rc = ov_open_callbacks (source, vf, NULL, 0, 
			       vorbisfile_callbacks)) != 0) {
    free (vf);
    if (rc == OV_ENOTVORBIS)
      return 1;
    
    if (rc == OV_EREAD)
      stream404_log_error ("Read error opening vorbis file");
    else if (rc == OV_EVERSION)
      stream404_log_error ("Vorbis version mismatch");
    else if (rc == OV_EBADHEADER)
      stream404_log_error ("Invalid vorbis header");
    else
      stream404_log_error ("Error in ov_open: %d", rc);
    
    return -1;
  }
  
  if (!(vorbis_data = (stream404_vorbis_in_t*)malloc (sizeof (stream404_vorbis_in_t)))) {
    stream404_log_error ("Malloc failed in stream404_vorbis_open");
    ov_clear (vf);
    free (vf);
    return -1;
  }

  if (!(vorbis_data->info = ov_info(vf, -1))) {
    stream404_log_error ("Vorbis: error reading vorbis info");
    stream404_vorbis_close (source);
    
    return -1;
  }
  
  if (vorbis_data->info->channels < 1) {
    stream404_log_error ("Vorbis: Cannot decode, %d channels of audio data!",
                    vorbis_data->info->channels);
    stream404_vorbis_close (source);
    
    return -1;
  }

  vorbis_data->vf = vf;
  vorbis_data->samples = 0;
  vorbis_data->link = -1;
  
  source->type = STREAM404_INPUT_VORBIS;
  source->data = vorbis_data;
  
  source->read = NULL;
  source->readpcm = stream404_vorbis_readpcm;
  source->close = http_vorbis_close; /* this has changed */
  
  in_vorbis_parse (source);
  
  return 0;
  
}

int
http_vorbis_close (input_stream_t* self)
{
  http_close(self);
  return stream404_vorbis_close(self);
}


/* --------------- MP3 PART --------------- */
typedef struct {
  unsigned char* buf;
  size_t len;
  int pos;
} stream404_mp3_in_t;

ssize_t http_mp3_read (input_stream_t* self, void* buf, size_t len);
int http_mp3_close (input_stream_t* self);

int stream404_mp3_parse (input_stream_t* source);
ssize_t stream404_mp3_read (input_stream_t* self, void* buf, size_t len);

ssize_t stream404_mp3_readpcm (input_stream_t* self, size_t len,
				 int16_t* left, int16_t* right);
int stream404_mp3_close (input_stream_t* self);


int http_open_mp3 (input_stream_t* self) 
{
  /* XXX this is absolutely temporar */
#define INPUT_BUFSIZ 4096
  char buf[INPUT_BUFSIZ];
  size_t len;

  stream404_mp3_in_t *mp3_data;
  int rc;

  if (http_open(self, METADATAON/*METADATAOFF*/) < 0) {
    stream404_log("cannot open remote URL\n");
    return -1;
  }
  
  if ((len = http_read (self, &buf, sizeof(char), INPUT_BUFSIZ)) <= 0) {
    stream404_log("Error reading header: %s", self->path);
    return -1;
  }
  
  if (!len)
    return -1;
  
  if (! ((mp3_data = (stream404_mp3_in_t*) malloc (sizeof (stream404_mp3_in_t))) &&
	 (mp3_data->buf = (unsigned char*) malloc (len))))
    {
       stream404_log_error ("Malloc failed in stream404_mp3_open");
      return -1;
    }
  
  memcpy (mp3_data->buf, buf, len);
  mp3_data->len = len;
  mp3_data->pos = 0;
  
  self->type = STREAM404_INPUT_MP3;
  self->data = mp3_data;

  self->read = http_mp3_read;
  
  self->readpcm = stream404_mp3_readpcm;
  /*
    self->readpcm = NULL;
  */
  self->close = http_mp3_close;
  
  stream404_id3v1_parse (self);
  
  if ((rc = stream404_mp3_parse (self))) {
    free (mp3_data->buf);
    free (mp3_data);
    return rc;
  }
  
  return 0;
}

/* input_stream_t wrapper for fread */
/* This is exactly the same as stream404_mp3_read except the read function */
ssize_t
http_mp3_read (input_stream_t* self, void* buf, size_t len)
{
  stream404_mp3_in_t* mp3_data = self->data;
  int remaining;
  int rlen = 0;

  if (mp3_data->buf) {
    remaining = mp3_data->len - mp3_data->pos;
    if (remaining > len) {
      rlen = len;
      memcpy (buf, mp3_data->buf + mp3_data->pos, len);
      mp3_data->pos += len;
    } else {
      rlen = remaining;
      memcpy (buf, mp3_data->buf + mp3_data->pos, remaining);
      free (mp3_data->buf);
      mp3_data->buf = NULL;
    }
  } else {
    /* we don't just use EOF because we'd like to avoid the ID3 tag */
    if (self->filesize && self->filesize - self->bytes_read < len)
      len = self->filesize - self->bytes_read;
    if (len)
      rlen = http_read(self, buf, 1, len);
  }

  self->bytes_read += rlen;

  return rlen;
}

int
http_mp3_close (input_stream_t* self)
{
  http_close(self);
  return stream404_mp3_close(self);
}

#ifdef HAVE_LIBFLAC
/* -------------------- FLAC PART ------------------------- */
/* -- data structures -- */
typedef struct {
  FLAC__StreamDecoder* decoder;
  unsigned char parsed;
  /* read buffer */
  char *buf;
  size_t len;
  /* write buffer */
  int16_t *left;
  int16_t *right;
  size_t olen;
} flac_in_t;

/* -- static prototypes -- */
extern int stream404_flac_readpcm (input_stream_t* self, size_t len,
			      int16_t* left, int16_t* right);
extern int stream404_flac_close (input_stream_t* self);

static FLAC__StreamDecoderReadStatus
flac_read_cb(const FLAC__StreamDecoder* decoder, FLAC__byte buffer[],
	     unsigned* bytes, void* client_data);
static FLAC__StreamDecoderWriteStatus
flac_write_cb(const FLAC__StreamDecoder* decoder, const FLAC__Frame* frame,
	      const FLAC__int32* const buffer[], void* client_data);
static void
flac_metadata_cb(const FLAC__StreamDecoder* decoder,
		 const FLAC__StreamMetadata* metadata, void* client_data);
static void
flac_error_cb(const FLAC__StreamDecoder* decoder,
	      FLAC__StreamDecoderErrorStatus status, void* client_data);

static inline int16_t scale16(int sample, int bps);

int http_open_flac(input_stream_t *);
int http_flac_close (input_stream_t*);

int http_open_flac(input_stream_t *source)
{
  flac_in_t* flac_data;
  FLAC__StreamDecoder* decoder;
#define INPUT_BUFSIZ 4096
  char buf[INPUT_BUFSIZ];
  size_t len;

  /* this is copied from stream404_open_vorbis XXX should be merged */
  if (http_open(source, METADATAOFF) < 0) {
    stream404_log("Cannot open remote URL");
    return -1;
  }

  if ((len = http_read (source, &buf, sizeof(char), INPUT_BUFSIZ)) <= 0) {
    stream404_log("Error reading header: %s", source->path);
    return -1;
  }
  
  if (!len)
    return -1;

  if (!(decoder = FLAC__stream_decoder_new())) {
    stream404_log_error("stream404_flac_open: Error allocating FLAC decoder");
    return -1;
  }

  FLAC__stream_decoder_set_read_callback(decoder, flac_read_cb);
  FLAC__stream_decoder_set_write_callback(decoder, flac_write_cb);
  FLAC__stream_decoder_set_metadata_callback(decoder, flac_metadata_cb);
  FLAC__stream_decoder_set_error_callback(decoder, flac_error_cb);

  FLAC__stream_decoder_set_metadata_respond(decoder, FLAC__METADATA_TYPE_VORBIS_COMMENT);

  if (!(flac_data = (flac_in_t*)malloc (sizeof (flac_in_t)))) {
    stream404_log_error ("Malloc failed in stream404_flac_open");
    goto errDecoder;
  }
  FLAC__stream_decoder_set_client_data(decoder, source);

  flac_data->decoder = decoder;
  flac_data->parsed = 0;
  flac_data->buf = buf;
  flac_data->len = len;

  source->data = flac_data;

  switch (FLAC__stream_decoder_init(decoder)) {
  case FLAC__STREAM_DECODER_SEARCH_FOR_METADATA:
    break;
  case FLAC__STREAM_DECODER_MEMORY_ALLOCATION_ERROR:
    stream404_log_error("Could not allocate memory during FLAC decoder init");
    goto errData;
  default:
    stream404_log_error("Unexpected error during FLAC decoder init");
    goto errData;
  }

  if (!FLAC__stream_decoder_process_single(decoder)) {
    stream404_log_error("Could not find FLAC metadata header");
    free(flac_data);
    FLAC__stream_decoder_delete(decoder);
    return 1;
  }
  if (!flac_data->parsed) {
    stream404_log_error("Could not find FLAC metadata header in prebuffer");
    free(flac_data);
    FLAC__stream_decoder_delete(decoder);
    return 1;
  }

  source->type = STREAM404_INPUT_FLAC;

  source->read = NULL;
  source->readpcm = stream404_flac_readpcm;
  source->close = http_flac_close;  /* this has changed */

  if (!FLAC__stream_decoder_process_until_end_of_metadata(decoder)) {
    stream404_log_error("Error seeking past FLAC metadata");
    goto errData;
  }
  
  return 0;

errData:
  free(flac_data);
errDecoder:
  FLAC__stream_decoder_delete(decoder);

  return -1;
}

int
http_flac_close (input_stream_t* self)
{
  http_close(self);
  return stream404_flac_close(self);
}

/* -------------------- FLAC Callbacks -------------------- */
/* -- callbacks -- */
static FLAC__StreamDecoderReadStatus
flac_read_cb(const FLAC__StreamDecoder* decoder, FLAC__byte buffer[],
	     unsigned* bytes, void* client_data)
{
  input_stream_t* self = (input_stream_t*)client_data;
  flac_in_t* flac_data = (flac_in_t*)self->data;
  ssize_t len;
  char errbuf[128];

  if (!flac_data->len) {
    if (!flac_data->parsed) {
      *bytes = 0;
      return FLAC__STREAM_DECODER_READ_STATUS_END_OF_STREAM;
    }
    if ((len = http_read(self, buffer, 1, *bytes)) > 0) {
	self->bytes_read += len;
	return FLAC__STREAM_DECODER_READ_STATUS_CONTINUE;
    }
    if (!len)
      return FLAC__STREAM_DECODER_READ_STATUS_END_OF_STREAM;
    stream404_log_error("Error reading FLAC stream: %s", stream404_util_strerror(errno, errbuf, sizeof(errbuf)));
    return FLAC__STREAM_DECODER_READ_STATUS_ABORT;
  }

  if (flac_data->len < *bytes)
    *bytes = flac_data->len;
  memcpy(buffer, flac_data->buf, *bytes);
  flac_data->len -= *bytes;

  return FLAC__STREAM_DECODER_READ_STATUS_CONTINUE;
}

static FLAC__StreamDecoderWriteStatus
flac_write_cb(const FLAC__StreamDecoder* decoder, const FLAC__Frame* frame,
	      const FLAC__int32* const buffer[], void* client_data)
{
  input_stream_t* self = (input_stream_t*)client_data;
  flac_in_t* flac_data = (flac_in_t*)self->data;
  int i;

  for (i = 0; i < frame->header.blocksize; i++) {
    flac_data->left[i] = scale16(buffer[0][i], frame->header.bits_per_sample);
    if (self->channels > 1) {
      flac_data->right[i] = scale16(buffer[1][i], frame->header.bits_per_sample);
    }
  }
  if (self->channels == 1)
    memcpy(flac_data->right, flac_data->left, frame->header.blocksize << 1);

  flac_data->olen = frame->header.blocksize;

  return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}

static void
flac_metadata_cb(const FLAC__StreamDecoder* decoder,
		 const FLAC__StreamMetadata* metadata, void* client_data)
{
  input_stream_t* self = (input_stream_t*)client_data;
  flac_in_t* flac_data = (flac_in_t*)self->data;
  FLAC__StreamMetadata_VorbisComment_Entry* comment;
  char *artist = NULL;
  char *title = NULL;
  int i;

  switch(metadata->type) {
  case FLAC__METADATA_TYPE_STREAMINFO:
    self->samplerate = metadata->data.stream_info.sample_rate;
    self->channels = metadata->data.stream_info.channels;
    /* calculate average bitrate 
       from fileinfo.c:226 in flac xmms plugin */
    self->bitrate = (int)(8.0 * (float)(self->filesize) / (1000.0 * (float)metadata->data.stream_info.total_samples / (float)metadata->data.stream_info.sample_rate));
    flac_data->parsed = 1;
    stream404_log_debug("Found FLAC file, %d Hz, %d channels, avg bitrate %d kbps", self->samplerate, self->channels, self->bitrate);
    break;
  case FLAC__METADATA_TYPE_VORBIS_COMMENT:
    for (i = 0; i < metadata->data.vorbis_comment.num_comments; i++) {
      comment = metadata->data.vorbis_comment.comments + i;
      
      stream404_log_debug ("FLAC (Vorbis) comment found: %s", comment->entry);

      if (comment->length >= 6 && !strncasecmp("artist", comment->entry, 6)) {
	if ((artist = malloc(comment->length - 6)))
	  snprintf(artist, comment->length - 6, "%s", comment->entry + 7);
      } else if (comment->length >= 5 && !strncasecmp("title", comment->entry, 5)) {
	if ((title = malloc(comment->length - 5)))
	  snprintf(title, comment->length - 5, "%s", comment->entry + 6);
      }
    }
    
    stream404_metadata_set(artist, title);
    if (artist)
      free(artist);
    if (title)
      free(title);
    break;
  default:
    stream404_log_debug("Ignoring unknown FLAC metadata type");
  }
}

static void
flac_error_cb(const FLAC__StreamDecoder* decoder,
	      FLAC__StreamDecoderErrorStatus status, void* client_data)
{
  switch (status) {
  case FLAC__STREAM_DECODER_ERROR_STATUS_LOST_SYNC:
    stream404_log_error("FLAC lost sync");
    break;
  case FLAC__STREAM_DECODER_ERROR_STATUS_BAD_HEADER:
    stream404_log_error("Bad header in FLAC stream");
    break;
  case FLAC__STREAM_DECODER_ERROR_STATUS_FRAME_CRC_MISMATCH:
    stream404_log_error("Bad CRC for FLAC frame");
    break;
  default:
    stream404_log_error("Unspecified error decoding FLAC stream");
  }
}

/* -- utility -- */
static inline int16_t scale16(int sample, int bps) {
  if (bps == 16)
    return sample;
  if (bps < 16)
    return sample << (16 - bps);
  return sample >> (bps - 16);
}

#endif
#ifdef HAVE_LIBFAAD
/* -------------------- MP4 PART ------------------------- */
/* XXX not supported !! */
#endif


int http_open_location(input_stream_t* source) 
{
  int ret = 1; 
  char *ext;
  
  /* always set it to 0 */
  private.control_metadata = 0;

  if (source->path[0] == '-' && source->path[1] == '\0') {
    stream404_log_debug("Reading audio from stdin ... now unavailable !");
    return -1;
  }

  /* to accelerate file reading */
  ext = strrchr(source->path, '.'); 
  if (ext) {
#ifdef HAVE_LIBFLAC
    if (!strcasecmp(ext, ".flac")) {
      if ((ret = http_open_flac(source)) < 0) {
	http_close(source);
	source->bytes_read = 0;
	return -1;
      }
      return 0;
    }
#endif

#ifdef HAVE_LIBFAAD
    /* peace of code from frederic fondriest */
    if (!strcasecmp(ext, ".mp4") ||	/* official extention */
	!strcasecmp(ext, ".m4a") ||	/* Apple mp4 extention */
	!strcasecmp(ext, ".aac")) {	/* old MPEG2/4-AAC extention */
      stream404_log("Reading remote HTTP MPEG4 file is not supported !");
      return -1;
    }
#endif  

    if (!strcasecmp(ext, ".ogg")) {
      if ((ret = http_open_vorbis(source)) < 0) {
	http_close(source);
	source->bytes_read = 0;
	return -1;
      }
      return 0;
    }

    if (!strcasecmp(ext, ".mp3")) {
      /* this is not a vorbis file, so test it as mp3 */
      if ((ret = http_open_mp3(source)) < 0) {
	http_close(source);
	source->bytes_read = 0;
	return -1;
      }
      return 0;
    }
  }    

  /* file has no explicit extension */
  /* second method look at the input_type_t */

#if 0 /* XXX */  
#ifdef HAVE_LIBFLAC
  if (source->type == STREAM404_INPUT_FLAC) {
    if ((ret = http_open_flac(source)) < 0) {
      http_close(source);
      source->bytes_read = 0;
      return -1;
    }
    return 0;
  }
#endif

#ifdef HAVE_LIBFAAD
  if (source->type == STREAM404_INPUT_MP4) {
    stream404_log("Reading remote HTTP MPEG4 file is not supported !");
    return -1;
  }
#endif  
#endif
  
  if (source->type == STREAM404_INPUT_VORBIS) {
    if ((ret = http_open_vorbis(source)) < 0) {
      http_close(source);
      source->bytes_read = 0;
      return -1;
    }
    return 0;
  }

  if (source->type == STREAM404_INPUT_MP3) {
    if ((ret = http_open_mp3(source)) < 0) {
      http_close(source);
      source->bytes_read = 0;
      return -1;
    }
    return 0;
  }

  stream404_log_debug("Remote file format is unknown, testing mp3 and ogg format (mp4 and flac are currently not supported)"); 
  
  
  if ((ret = http_open_vorbis(source)) < 0)
    return -1;
  if (ret == 1) {
    http_close(source);
    source->bytes_read = 0;
    if ((ret = http_open_mp3(source)) < 0)
      return -1;
  }
  
  stream404_log_error("Remote file format is unknown");

  return -1;
}


size_t write_callback (void *ptr, size_t size, size_t nmemb, void *arg)
{
  char *newbuff;
  int rembuff;

  http_private_t *myarg = arg;

  size *= nmemb;

  /* remaining space in buffer */
  rembuff = myarg->buffer_len - myarg->buffer_pos;
  
  if (size > rembuff) {
    /* not enuf space in buffer */
    newbuff = realloc(myarg->buffer,myarg->buffer_len + (size - rembuff));
    if(newbuff == NULL) {
      //fprintf(stderr,"callback buffer grow failed\n");
      size = rembuff;
    } else {
      /* realloc suceeded increase buffer size*/
      myarg->buffer_len+=size - rembuff;
      myarg->buffer=newbuff;
      
      /*printf("Callback buffer grown to %d bytes\n",url->buffer_len);*/
    }
  }
  
  memcpy(&myarg->buffer[myarg->buffer_pos], ptr, size);
  myarg->buffer_pos += size;
  
  /*fprintf(stderr, "callback %d size bytes\n", size);*/
  
  return size;
}


size_t header_callback(void *ptr, size_t size, size_t nmemb, void *arg)
{
  http_private_t *myarg = arg;

  if (!strncmp(ptr, "icy-metaint:", 12)) {
    myarg->icy_metaint = atoi(ptr + 12);
    //fprintf(stdout, "icy metadata = %d\n", myarg->icy_metaint);
  }

  return size*nmemb;
}


/* use to attempt to fill the read buffer up to requested number of bytes */
static int
curl_fill_buffer(http_private_t *private, int want, int waittime)
{
  fd_set fdread;
  fd_set fdwrite;
  fd_set fdexcep;
  int maxfd;
  struct timeval timeout;
  int rc;
  
  /* only attempt to fill buffer if transactions still running and buffer
   * doesnt exceed required size already
   */
  if((!private->still_running) || (private->buffer_pos > want))
    return 0;
  
    /* attempt to fill buffer */
    do {
      FD_ZERO(&fdread);
      FD_ZERO(&fdwrite);
      FD_ZERO(&fdexcep);
      
      /* set a suitable timeout to fail on */
      timeout.tv_sec = 5; /* 3 seconds */
      timeout.tv_usec = 0;
      
      /* get file descriptors from the transfers */
      curl_multi_fdset(private->multi_handle, &fdread, &fdwrite, 
		       &fdexcep, &maxfd);
      
      rc = select(maxfd+1, &fdread, &fdwrite, &fdexcep, &timeout);
      
      switch(rc) {
      case -1:
	/* select error */
	/* select may be interrupted by a signal (for example SIGPOLL
	   to change metadata) if so then continue */ 
	if (errno != EINTR) {
	  stream404_log_debug("Cannot read anymore the current http file/stream");
	  private->buffer_pos = 0; /* explicit stop */
	  return 0;
	}
	break;
	
      case 0:
	stream404_log_debug("Timeout in reading the current http file/stream");
	private->buffer_pos = 0; /* explicit stop */
	return 0;
	break;
	
      default:
	/* timeout or readable/writable sockets */
	/* note we *could* be more efficient and not wait for
	 * CURLM_CALL_MULTI_PERFORM to clear here and check it on re-entry
	 * but that gets messy */
	while(curl_multi_perform(private->multi_handle, 
				 &private->still_running) ==
	      CURLM_CALL_MULTI_PERFORM);
	
	break;
      }
    } while(private->still_running && (private->buffer_pos < want));
    return 1;
}

/* use to remove want bytes from the front of a files buffer */
static int
curl_use_buffer(http_private_t *private, int want)
{
  /* sort out buffer */
  if((private->buffer_pos - want) <=0) {
    /* ditch buffer - write will recreate */
    if(private->buffer)
      free(private->buffer);
    
    private->buffer=NULL;
    private->buffer_pos=0;
    private->buffer_len=0;
  }
  else {
    /* move rest down make it available for later */
    memmove(private->buffer,
	    &private->buffer[want],
	    (private->buffer_pos - want));
    
    private->buffer_pos -= want;
  }
  return 0;
}

int curl_get_data(http_private_t *private, void *ptr, int want)
{
  //memset(ptr, 0, want);
  curl_fill_buffer(private, want, 1);
  /* check if theres data in the buffer - if not curl_fill_buffer()
   * either errored or EOF */
  if(!private->buffer_pos)
    return 0;
  
  /* ensure only available data is considered */
  if(private->buffer_pos < want)
    want = private->buffer_pos;
  
  /* xfer data to caller */
  memcpy(ptr, private->buffer, want);
  
  curl_use_buffer(private, want);
  
  return want;
}

void clean_curl_opts(http_private_t *private)
{
  if (private->curl_handle != NULL && private->multi_handle != NULL)
    curl_multi_remove_handle(private->multi_handle, private->curl_handle);
  if (private->curl_handle != NULL)
    curl_easy_cleanup(private->curl_handle);
  if (private->header_list != NULL)
    curl_slist_free_all(private->header_list);
  if (private->httpaliases_list != NULL)
    curl_slist_free_all(private->httpaliases_list);
}

void set_curl_opts (input_stream_t *input, http_private_t *private, int enable)
{
  CURL *handle; 
  
  private->input = input;
  private->curl_handle = NULL;
  private->header_list = NULL;
  private->httpaliases_list = NULL;
  private->icy_metaint = 0;
  private->bytes_read = 0;
  
  private->header_list = curl_slist_append(private->header_list, "Accept: application/ogg, audio/x-flac, audio/mpeg");
  if (enable == METADATAON)
    private->header_list = curl_slist_append(private->header_list, "Icy-MetaData:1");

  private->httpaliases_list = curl_slist_append(private->httpaliases_list, 
					       "HTTP/1.0 200 OK");
  private->httpaliases_list = curl_slist_append(private->httpaliases_list, 
					       "HTTP/1.1 200 OK");
  private->httpaliases_list = curl_slist_append(private->httpaliases_list, 
					       "ICY 200 OK");
  

  if ((private->header_list == NULL) || (private->httpaliases_list == NULL)) {
    clean_curl_opts(private);
    return;
  }

  handle = private->curl_handle = curl_easy_init();
  if (handle == NULL) {
    clean_curl_opts(private);
    return;
  }

  
  curl_easy_setopt(handle, CURLOPT_FILE, private);
  curl_easy_setopt(handle, CURLOPT_WRITEFUNCTION, write_callback);
  curl_easy_setopt(handle, CURLOPT_URL, private->input->path);
  /* curl_easy_setopt(handle, CURLOPT_MUTE, 1); */
  curl_easy_setopt(handle, CURLOPT_ERRORBUFFER, private->error);
  curl_easy_setopt(handle, CURLOPT_HEADERFUNCTION, header_callback);
  curl_easy_setopt(handle, CURLOPT_NOPROGRESS, 1);
  curl_easy_setopt(handle, CURLOPT_USERAGENT, "stream404/0.2");
  curl_easy_setopt(handle, CURLOPT_HTTPHEADER, private->header_list);
  curl_easy_setopt(handle, CURLOPT_WRITEHEADER, private);
  curl_easy_setopt(handle, CURLOPT_HTTP200ALIASES, private->httpaliases_list);
  curl_easy_setopt(handle, CURLOPT_FOLLOWLOCATION, 1);
  /* 5 seconds max to wait before considering the stream as dead */
  curl_easy_setopt(handle, CURLOPT_CONNECTTIMEOUT, 5); 

  if(!private->multi_handle)
    private->multi_handle = curl_multi_init();
  
  curl_multi_add_handle(private->multi_handle, handle);
  
}


int http_open (input_stream_t *input, int enable)
{
  set_curl_opts(input, &private, enable);
  
  /* lets start the fetch */
  while(curl_multi_perform(private.multi_handle, &private.still_running) ==
	CURLM_CALL_MULTI_PERFORM );
  
  if((private.buffer_pos == 0) && (!private.still_running)) {
    /* if still_running is 0 now, we should return NULL */
    clean_curl_opts(&private);
    return -1;
  }
  return 0;
}


char *subnstr_until(const char *str, char *until, char *newstr, int maxlen)
{
  const char *p = str;
  int len = 0;
  
  for(len = 0; strncmp(p, until, strlen(until)) != 0 && len < maxlen; p++) {
    newstr[len] = *p;
    len++;
  }
  newstr[len] = '\0'; 
  return newstr;
}

/* this is the new version : what changes is when we read the metadata, we try
   instead of returning a buffer with less value read, try to fill it by
   reading another amount of data
*/
int http_read (input_stream_t *input, void *ptr, size_t size, size_t nmemb)
{
  int i = 0;
  int bytes_read;
  char streamtitle[4096];

  if (input == NULL || ptr == NULL) {
    stream404_log_debug ("http_read : Either input or ptr is NULL !!");
    return 0;
  }
 
  if ((bytes_read = curl_get_data(&private, ptr, size * nmemb)) == 0)
    return 0;

  if (private.icy_metaint == 0) {
    /* or we didn't wanted icy_metaint or we listen at an ogg vorbis stream*/
    private.bytes_read += bytes_read;
    return bytes_read;
  }

  /* This part of the code concerns only mp3 streams !!!! */
  if (private.bytes_read != 0) {
    if (private.icy_metaint <= bytes_read + private.bytes_read) {
      /* have we read enough ? */
      if (private.bytes_read % private.icy_metaint != 0) {
	i = private.icy_metaint - 
	  private.bytes_read % private.icy_metaint;
      } else {
	/* the problem is if the test before i was equal to private.icy_metaint
	   then the next try it will be equal to 0, which means we read twice
	   the buffer instead of waiting for a  "icy_metaint" delay */
	if ((private.bytes_read + private.icy_metaint) % private.icy_metaint == 0)
	  i = private.icy_metaint;
	/* else i = 0 (implicit)*/
      }
    } else {
      /* a -1 value means we stop */
      i = -1;
    }
  } else {
    /* in this case we never read anything so i cannot be equal to 0 */
    if (private.icy_metaint <= bytes_read + private.bytes_read) {
      /* have we read enough ? */
      i = private.icy_metaint;
    } else {
      /* a -1 value means we stop */
      i = -1;
    }
  }

  if (i != -1) {
    /* there could be more than one metaint inside the current read buffer */
    while (i <= bytes_read) {
      //fprintf(stdout, "i = %d private.bytes_read = %d", i, private.bytes_read);
      
      if (i == bytes_read) {
	char d;
	/* if we can't read the next byte we stop */
	if (curl_get_data(&private, &d, 1) != 1) {
	  //fprintf(stdout, "error reading another byte!!!\n");
	  /* break at the end */
	} else {
	  /* what did we read ? */
	  //fprintf(stdout, "value of offset (i == bytes_read) = %d", (int)d);
	  switch (d) {
	  case 0: 
	    {
	      //fprintf(stdout, " so do nothing\n");
	    }
	    break;
	  default:
	    {
	      int how = d * 16;
	      char cr[4096];
	      if (curl_get_data(&private, cr, how) != how)
		return 0;
	      //curl_get_data(&private, cr, how);
	      //fprintf(stdout, "then meta data (offset = 4096) = %s\n\n", cr);
	      stream404_metadata_set(NULL, subnstr_until(cr + strlen("StreamTitle='"), "';", streamtitle, 4096));

	      /* for every ogg stream flush and set the new metadata */
	      if (private.control_metadata > 0) {
		stream404_stream_t* stream;
		stream_flush_vorbis();
		for (stream = stream404_config.streams; stream; 
		     stream = stream->next) {
		  if (stream->type == STREAM404_OUTPUT_VORBIS) { 
		    vorbis_reencode_reset(input, stream);
		  }
		}
	      } else {
		/* we cannot ask vorbis to flush so early */
		private.control_metadata++;
	      }
	      stream404_metadata_update (0);
	    }
	    break;
	  } 
	}
	/* return values are set at the end */
	break; /* to stop the loop */
      } /* if (i == bytes_read) */
      else {
	char *c;
	c = ptr + i;
	//fprintf(stdout, "value of offset = %d   (offset -1 %d) (offset + 1 %d)\n", (int)*c, (int)*(c-1), (int)*(c+1));
	switch (*c) {
	case 0: 
	  {
	    int j;
	    int d;
	    for (j = i; j < bytes_read; j++) {
	      if (j <= bytes_read - 1) 
		*((char *)ptr + j) = *((char *)ptr + j + 1);
	    }
	    /* try to read one more time char */
	    if (curl_get_data(&private, &d, 1) != 1) {
	      /* we couldn't so we tell we have read bytes_read minus one byte */
	      bytes_read = bytes_read - 1;
	    } else {
	      /* okay we could read than byte, put it in the buffer */
	      *((char *)ptr + (bytes_read - 1)) = (char) d;
	      /* we keep we read bytes_read */
	    }
	    //bytes_read = bytes_read - 1;
	    /* then add the next bytes to be read */
	    i += private.icy_metaint;
	  }
	  break;
	default: 
	  {
	    char br[4096];
	    int j;
	    int how = *c * 16;
	    if (how + i < bytes_read) {
	      strncpy(br, (char *)ptr + i + 1, how);
	      //stream404_log_debug("meta data = %s", br);
	      stream404_metadata_set(NULL, subnstr_until(br + strlen("StreamTitle='"), "';", streamtitle, 4096));
	      
	      /* for every ogg stream flush and set the new metadata */
	      if (private.control_metadata > 0) {
		stream404_stream_t* stream;
		stream_flush_vorbis();
		for (stream = stream404_config.streams; stream; 
		     stream = stream->next) {
		  if (stream->type == STREAM404_OUTPUT_VORBIS) { 
		    vorbis_reencode_reset(input, stream);
		  }
		}
	      } else {
		/* we cannot ask vorbis to flush so early */
		private.control_metadata++;
	      }

	      stream404_metadata_update (0);
	      //fprintf(stderr, "buffer before \n------------------\n%s\n----------\n", (char *)ptr);
	      //write(2, "buffer before \n------------------\n%s\n----------\n", sizeof("buffer before \n------------------\n%s\n----------\n"));
	      //write(2, ptr, bytes_read);
	      for (j = i; j < bytes_read; j++) {
		/* XXX if (j <= bytes_read - how) XXX */ 
		if (j < bytes_read - how) 
		  *((char *)ptr + j) = *((char *)ptr + j + how + 1); /* XXX added + 1 */
	      }
	      //fprintf(stderr, "buffer after \n------------------\n%s\n----------\n", (char *)ptr);
	      
	      bytes_read = bytes_read - how - 1;
	      i += private.icy_metaint + 1;
	    } else {
	      char s[4096], br2[4096];
	      int nb;
	      //stream404_log_debug ("we haven't read enough how + i = %d and we read bytes_read = %d , we need to read %d bytes more !!!!", how + i, bytes_read, how + i - bytes_read + 1);
	      //s = calloc(how + offset - bytes_read, sizeof(char));
	      
	      nb = curl_get_data(&private, s, how + i - bytes_read + 1);
	      if (nb <= 0)
		return 0;
	      //fprintf(stdout, "we have read nb = %d, value = %s\n\n\n\n\n\n", nb, s);
	      if (i != bytes_read) {
		strncpy(br2, (char *)ptr + i + 1, bytes_read - i - 1);
		//fprintf(stdout, "first meta data part = %s\n\n", br2);
	      }
	      strncat(br2 + bytes_read - i - 1, s, nb);
	      //fprintf(stdout, "\n\nmeta data = %s\n\n", br2);
	      //stream404_log_debug("meta data = %s", br2);
	      stream404_metadata_set(NULL, subnstr_until(br2 + strlen("StreamTitle='"), "';", streamtitle, 4096));

	      /* for every ogg stream flush and set the new metadata */
	      if (private.control_metadata > 0) {
		stream404_stream_t* stream;
		stream_flush_vorbis();
		for (stream = stream404_config.streams; stream; 
		     stream = stream->next) {
		  if (stream->type == STREAM404_OUTPUT_VORBIS) { 
		    vorbis_reencode_reset(input, stream);
		  }
		}
	      } else {
		/* we cannot ask vorbis to flush so early */
		private.control_metadata++;
	      }

	      stream404_metadata_update (0);
	      bytes_read = i;
	      i += private.icy_metaint + 1;
	    }
	    break;
	  }
	}
      } /* else (if (i != bytes_read)) */
      
    } /* while (i <= bytes_read) */
  } /* if (i != -1) */

  

  private.bytes_read += bytes_read;
  //fprintf(stdout, "bytes_read = %d, private.bytes_read = %d, metadata = %d\n", bytes_read, private.bytes_read, private.icy_metaint);
  return bytes_read;
}


int http_seek (input_stream_t *input, long offset, int whence)
{
  return -1;
}


long http_tell (input_stream_t *input)
{
  return 0;
}


void http_close (input_stream_t *input)
{
  clean_curl_opts(&private);
  private.bytes_read = 0;
  if(private.buffer)
    free(private.buffer);
  
  private.buffer=NULL;
  private.buffer_pos=0;
  private.buffer_len=0;
}
