/* $Id: stream.c,v 1.3 2005/06/24 22:01:51 bedlam Exp $ */
/* stream.c
 * - Functions for streaming in ices
 * Copyright (c) 2000 Alexander Havng
 * Copyright (c) 2001-4 Brendan Cully
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include "definitions.h"
#include "metadata.h"

#ifdef HAVE_LIBVORBISFILE
#include "in_vorbis.h"
#endif
#ifdef HAVE_LIBFAAD
#include "in_mp4.h"
#endif
#ifdef HAVE_LIBFLAC
#include "in_flac.h"
#endif

#ifdef TIME_WITH_SYS_TIME
#  include <sys/time.h>
#  include <time.h>
#else
#  ifdef HAVE_SYS_TIME_H
#    include <sys/time.h>
#  else
#    include <time.h>
#  endif
#endif

//#define INPUT_BUFSIZ 4096
#define INPUT_BUFSIZ 2048
//#define INPUT_BUFSIZ 512

#define OUTPUT_BUFSIZ 32768
/* sleep this long in ms when every stream has errors */
#define ERROR_DELAY 999

#define SET_DEFAULT_SOURCE src.type = STREAM404_INPUT_UNKNOWN; \
	                   src.location = LOCAL;          \
                           src.spec = _FILE;              \


typedef struct {
  char* data;
  size_t len;
} buffer_t;

static volatile int finish_send = 0;
static volatile int finish_urgently = 0;
static int needtoupdatemetadata = 0;

/* Private function declarations */
static int stream_connect (stream404_stream_t* stream);
static int stream_send (stream404_config_t* config, input_stream_t* source);
static int stream_send_data (stream404_stream_t* stream, unsigned char* buf,
			     size_t len);
static int stream_open_source (input_stream_t* source);
static int stream_needs_reencoding (input_stream_t* source, stream404_stream_t* stream);
void stream404_stream_need2updatemetadata(void);
void timing_sleep(int sleeptime);
/* XXX obsoleted by timing_sleep 
   static int stream_get_sync(stream404_config_t* config);
*/

void init_stream(input_stream_t *input)
{
  input->type = input->spec = input->location = input->reconnects = 0;
  input->path = NULL;
}

/* Public function definitions */
input_stream_t src; /* source is declared outside */

/* Top level streaming function, called once from main() to
 * connect to server and start streaming */
void
stream404_stream_loop (stream404_config_t* config)
{
  int consecutive_errors = 0;
  //input_stream_t source;
  stream404_stream_t* stream;
  int rc;

  init_stream(&src);
  setproctitle("stream404", "main stream parent");

  while (1) {
    
    /* if finish_urgently is set handle it, whatever we play 
       (file or stream) */
    if ((finish_urgently == 1)) {
      /* these three values may be set by the stream404_get_next function */
      SET_DEFAULT_SOURCE;

      if ((src.path = stream404_playlist_get_next_urg (&src)) == NULL) {
	stream404_log("There is no file to be played in urgent table, so look at the normal one");
	/* should not be mandatory */
	SET_DEFAULT_SOURCE;
	src.path = stream404_playlist_get_next (&src);
      }
    } else {
      /* we try to know why we loop until here 
	 2 cases : 
	 we were asked to play a new file so finish_send is set to 1
	 else 
	 - the current file finished
	 - the stream dropped (a stream is identified by the source.spec)

	 if a stream has been droped we need to look a its reconnect time 
	 which is set in a sql argument 
       */
      if (finish_send == 0 && src.reconnects > 0 && src.spec == STREAM) {
	src.reconnects--;
	/* XXX to be cleaned there is a problem */
	/* don't need to set the source.path as it is already filled 
	   and never call stream404_util_free
	 */
      } else {
	SET_DEFAULT_SOURCE;
	src.path = stream404_playlist_get_next (&src);
      }
    }

    if (!(src.path && src.path[0])) {
      stream404_log ("Playlist file name is empty, shutting down.");
      stream404_setup_shutdown ();
    }

    stream404_cue_set_lineno (stream404_playlist_get_current_lineno ());

    stream404_metadata_set (NULL, NULL);
    stream404_metadata_set_file (src.path);

    /* This stops stream404 from entering a loop with 10-20 lines of output per
	 second. Usually caused by a playlist handler that produces only
	 invalid file names. */
    if (consecutive_errors > 15) {
      stream404_log ("Exiting after 15 consecutive errors.");
      stream404_util_free (src.path);
      stream404_setup_shutdown ();
    }

    if (stream_open_source (&src) < 0) {
      stream404_log ("Error opening %s: %s", src.path, stream404_log_get_error ());
      /* a stream can be unreachable for a moment without beeing a serious
	 error, that's why there we set the src.reconnects 
	 but if it is a file then clean the path  
      */
      if (src.spec != STREAM) {
	stream404_util_free (src.path);
	consecutive_errors++;
      }
      continue;
    }

    if (!src.read)
      for (stream = config->streams; stream; stream = stream->next)
        if (!stream->reencode) {
          stream404_log ("Cannot play %s without reencoding", src.path);
          src.close (&src);
          stream404_util_free (src.path);
          consecutive_errors++;
          continue;
	}

    rc = stream_send (config, &src);
    src.close (&src);

    /* If something goes on while transfering, we just go on */
    if (rc < 0) {
      stream404_log ("Encountered error while transfering %s: %s", src.path, stream404_log_get_error ());

      /* a stream can be unreachable for a moment without beeing a serious
	 error, that's why there we set the src.reconnects 
	 but if it is a file then clean the path  
      */
      if (src.spec != STREAM) {
	stream404_util_free (src.path);
	consecutive_errors++;
      }
      continue;
    } else {
      /* Reset the consecutive error counter */
      consecutive_errors = 0;
    }

    if (finish_urgently == 1 || finish_send == 1 || src.reconnects == 0 || src.spec != STREAM) {
      stream404_util_free (src.path);
    }
  }
}

/* set a flag to tell stream_send to finish early */
void
stream404_stream_next (void)
{
  finish_send = 1;
  /* we received a signal so we are sure not 
     to continue reading the same file */
  src.reconnects = 0;
}

/* set a flag to tell stream_send to finish early and urgently */
void
stream404_stream_urgent_next (void)
{
  finish_urgently = 1;
  stream404_stream_next ();
}

/* set a flag to set a new metadata */
void 
stream404_stream_need2updatemetadata (void)
{
    needtoupdatemetadata = 1;
}

/* This function is called to stream a single file */
static int
stream_send (stream404_config_t* config, input_stream_t* source)
{
  stream404_stream_t* stream;
  unsigned char ibuf[INPUT_BUFSIZ];
  char namespace[1024];
  ssize_t len;
  ssize_t olen;
  int samples;
  int rc;
  int do_sleep;
  int decode = 0;
  buffer_t obuf;
  stream404_plugin_t *plugin;
  /* worst case decode: 22050 Hz at 8kbs = 44.1 samples/byte */
  static int16_t left[INPUT_BUFSIZ * 45];
  static int16_t right[INPUT_BUFSIZ * 45];

  static int16_t lr[INPUT_BUFSIZ * 45 * 2 + 1];

  ogg_page og;

  static int16_t* rightp;

  obuf.data = NULL;
  obuf.len = 0;

  if (config->reencode) {
    stream404_reencode_reset (source);
    if (config->plugins) {
      decode = 1;
      for (plugin = config->plugins; plugin; plugin = plugin->next)
	plugin->new_track(source);
    } else
      for (stream = config->streams; stream; stream = stream->next)
	if (stream->reencode && stream_needs_reencoding (source, stream)) {
	  decode = 1;
	  break;
	}
  }

  /* XXXXXXXXXXXXX always set decode to yes */
  decode = 1;

  if (decode) {
    obuf.len = OUTPUT_BUFSIZ;
    if (!(obuf.data = malloc(OUTPUT_BUFSIZ))) {
      stream404_log_error("Error allocating encode buffer");
      return -1;
    }
  }

  for (stream = config->streams; stream; stream = stream->next)
    stream->errs = 0;
  
  stream404_log ("Playing %s", source->path);

  stream404_metadata_update (0);

  finish_send = 0;
  finish_urgently = 0;
  while (! finish_send) {

    /* check if we need to update metadata */
    if (needtoupdatemetadata) {
      char *meta;
      stream404_log("getting new metadata");
      if ((meta = stream404_playlist_get_metadata()) != NULL) {
	stream404_metadata_set(NULL, meta);
	{
	  stream404_stream_t* stream;
	  stream_flush_vorbis();
	  for (stream = config->streams; stream; 
	       stream = stream->next) {
	    if (stream->type == STREAM404_OUTPUT_VORBIS) { 
	      vorbis_reencode_reset(source, stream);
	    }
	  }
	}
	stream404_util_free(meta);
	stream404_metadata_update (0);
      }
      needtoupdatemetadata = 0;
    }

    len = samples = 0;

    /* fetch input buffer 
    if (source->read) {
      len = source->read (source, ibuf, sizeof (ibuf));
      if (decode)
        samples = stream404_reencode_decode (ibuf, len, sizeof (left), left, right);
    } else if (source->readpcm) {
    */
    /*
      len = samples = source->readpcm (source, sizeof (left), left, right);
     */
      if ((len = samples = 
	  source->readpcm (source, sizeof (left), left, right)) > INPUT_BUFSIZ * 45)
	{
	  /*
	  stream404_log_debug("too much samples, skipping ...");
	  continue;
	  */
	  stream404_log_debug("too much samples, setting to max value %d ...", INPUT_BUFSIZ * 45);
	  len = samples = INPUT_BUFSIZ * 45;

	}
      /*
	}*/

#ifdef HAVE_LIBLAME
      /* run output through plugin */
      for (plugin = config->plugins; plugin; plugin = plugin->next)
	if (samples)
	  samples = plugin->process(samples, left, right);
#endif

    if (len == 0) {
      stream404_log_debug ("Done sending");
      break;
    }
    if (len < 0) {
      stream404_log_error ("Read error: %s", stream404_util_strerror (errno, namespace, 1024));
      goto err;
    }

    /* fprintf(stderr, "samples = %d, len = %d\n", samples, len); */
    
    /* before begining to encode/send we need to build the buffer for 
       the ogg/vorbis stream */
    {
      int i;
      if (source->channels == 1) {
	for (i = 0; i < samples; i++) {
	  
	  lr[2*i] = left[i];
	 
	  lr[2*i + 1] = left[i];
	 
	}
      } else {
	for (i = 0; i < samples; i++) {
	  lr[2*i] = left[i];
	 
	  lr[2*i + 1] = right[i];
	 
	}
      }
    }
    
    if (samples == 0) {
      /* samples == 0 and len != 0 which means we are at the begining 
	 of a file and crossfade is buffering
      */
      continue;
    }

    do_sleep = 1;
    while (do_sleep) {
      rc = olen = 0;

      /* we should sleep before sending anything and we should do it there */

      for (stream = config->streams; stream; stream = stream->next) {
	switch (stream->type) {
	case STREAM404_OUTPUT_VORBIS: {
	  vorbis_encoder_t *enc = (vorbis_encoder_t *)stream->encoder_state;
	  resample_state_t *resamp = (resample_state_t *)stream->resample_state;
	  downmix_state_t *downmix = (downmix_state_t *)stream->downmix_state;

	  /* downmix ?! */
	  if (source->channels == 2 && stream->out_numchannels == 1) {
	    downmix_buffer(downmix, (signed char *)lr, samples*2*source->channels, 0);
	    resample_buffer_float(resamp, (float **)downmix->buffer, samples*2*source->channels/4);
	    vorbis_encode_data_float(enc, resamp->buffers, resamp->buffill);
	  } else {
	    if (source->channels == 1)
	      samples = samples*2;
	    resample_buffer(resamp, (signed char *)lr, samples*2*source->channels, 0);
            vorbis_encode_data_float(enc, resamp->buffers, resamp->buffill);
	  }

	  while (vorbis_encode_dataout(enc, &og) > 0) {
            if ((rc = stream_send_data(stream, og.header, og.header_len)) 
		< 0) {
	      /* This stream can't restart until the next
	       * logical stream comes along, since the
	       * server won't have any cached headers for
	       * this source/connection. So, don't continue
	       * yet.
	       */
	     
	      /* clear resample buffer too */
	      resample_clear(resamp);
	      vorbis_encode_clear(enc);
	      vorbis_reencode_reset (source, stream);
	      break;
	    }
            if ((rc = stream_send_data(stream, og.body, og.body_len)) < 0) {
	      /* This stream can't restart until the next
	       * logical stream comes along, since the
	       * server won't have any cached headers for
	       * this source/connection. So, don't continue
	       * yet.
	       */
	    
	      /* clear resample buffer too */
	      resample_clear(resamp);
	      vorbis_encode_clear(enc);
	      vorbis_reencode_reset (source, stream);
	      break;
	    }
	  }
	}
	  break;


	case STREAM404_OUTPUT_MP3:
	default:
	   /* don't reencode if the source is MP3 and the same bitrate */
	  if (stream->reencode && (config->plugins || stream_needs_reencoding (source, stream))) {

	    if (samples > 0) {
	      /* for some reason we have to manually duplicate right from left to get
	       * LAME to output stereo from a mono source */
	      if (source->channels == 1 && stream->out_numchannels != 1)
		rightp = left;
	      else
		rightp = right;
	      if (obuf.len < (unsigned int) (7200 + samples + samples / 4)) {
		char *tmpbuf;
		
		/* pessimistic estimate from lame.h */
		obuf.len = 7200 + 5 * samples / 2;
		if (!(tmpbuf = realloc(obuf.data, obuf.len))) {
		  stream404_log_error ("Error growing output buffer, aborting track");
		  goto err;
		}
		obuf.data = tmpbuf;
		stream404_log_debug ("Grew output buffer to %d bytes", obuf.len);
	      }
	      if ((olen = stream404_reencode (stream, samples, left, rightp, obuf.data,
					 obuf.len)) < -1) {
		stream404_log_error ("Reencoding error, aborting track");
		goto err;
	      } else if (olen == -1) {
		char *tmpbuf;
		
		if ((tmpbuf = realloc(obuf.data, obuf.len + OUTPUT_BUFSIZ))) {
                  obuf.data = tmpbuf;
		  obuf.len += OUTPUT_BUFSIZ;
		  stream404_log_debug ("Grew output buffer to %d bytes", obuf.len);
		} else
		  stream404_log_debug ("%d byte output buffer is too small", obuf.len);
	      } else if (olen > 0) {
		rc = stream_send_data (stream, obuf.data, olen);
	      }
	    }
	  } else
	    rc = stream_send_data (stream, ibuf, len);
	}
        

        if (rc < 0) {
          if (stream->errs > 10) {
            stream404_log ("10 server errors occured");
	    stream->errs = 0;
	    
	    /* 
	       stream404_log ("Too many stream errors, giving up");
	       stream404_setup_shutdown ();
	    */
          }
          stream404_log ("Error during send: %s", stream404_log_get_error ());
        } else {
	      do_sleep = 0;
        }
        /* this is so if we have errors on every stream we pause before
         * attempting to reconnect */
        if (do_sleep) {
          struct timeval delay;
          delay.tv_sec = ERROR_DELAY / 1000;
          delay.tv_usec = ERROR_DELAY % 1000 * 1000;

          select (1, NULL, NULL, NULL, &delay);
        }

      } /* for (stream = config->streams; stream; stream = stream->next) */
    }
    stream404_cue_update (source);
  } /* while (! finish_send) */


  /* we should sleep before sending anything and we should do it there */

  for (stream = config->streams; stream; stream = stream->next)
    if (stream->reencode && stream_needs_reencoding (source, stream)) {
      switch (stream->type) {
      case STREAM404_OUTPUT_VORBIS: {
	  vorbis_encoder_t *enc = (vorbis_encoder_t *)stream->encoder_state;
	  resample_state_t *resamp = (resample_state_t *)stream->resample_state;

	  resample_finish(resamp);
	  vorbis_encode_data_float(enc, resamp->buffers, resamp->buffill);
	  resample_clear(resamp);
	  /* set in  vorbis_reencode_reset 
	     resamp = resample_initialise (stream->channels,
	     sdsc->stream->resampleinrate, sdsc->stream->resampleoutrate);
	  */
	  vorbis_encode_finish(enc);

	  stream404_log_debug("flushing current stream");
	  while(vorbis_encode_flush(enc, &og) != 0) {
	    if ((rc = stream_send_data(stream, og.header, og.header_len)) < 0) 
	      break;
	    if ((rc = stream_send_data(stream, og.body, og.body_len)) < 0)
	      break;
	  }
	  vorbis_encode_clear(enc);
	break;
      }
      case STREAM404_OUTPUT_MP3: 
      default : {
	len = stream404_reencode_flush (stream, obuf.data, obuf.len);
	if (len > 0)
	  rc = shout_send (stream->conn, obuf.data, len);
	break;
      }
      }
    }

  if (obuf.data)
    free(obuf.data);
  
  return 0;

err:
  if (obuf.data)
    free(obuf.data);
  return -1;
}

/* this function flush the encoder and sends last bits to the server
   it should be called as soon as there is a change in a vorbis/mp3 stream */
extern stream404_config_t stream404_config;
void stream_flush_vorbis()
{
  ogg_page og;
  stream404_stream_t* stream;
  int rc;
  for (stream = stream404_config.streams; stream; stream = stream->next)
    if (stream->type == STREAM404_OUTPUT_VORBIS) { 
      vorbis_encoder_t *enc = (vorbis_encoder_t *)stream->encoder_state;
      resample_state_t *resamp = (resample_state_t *)stream->resample_state;

      resample_finish(resamp);
      vorbis_encode_data_float(enc, resamp->buffers, resamp->buffill);
      resample_clear(resamp);

      vorbis_encode_finish(enc);
      while(vorbis_encode_flush(enc, &og) != 0) {
	if ((rc = stream_send_data(stream, og.header, og.header_len)) < 0)
	  break;
	if ((rc = stream_send_data(stream, og.body, og.body_len)) < 0)
	  break;
      }
      vorbis_encode_clear(enc);
    }
}

/* open up path, figure out what kind of input it is, and set up source */
static int
stream_open_source (input_stream_t* source)
{
  char buf[INPUT_BUFSIZ], *ext;
  size_t len;
  int fd;
  int rc;

  source->filesize = 0;
  source->bytes_read = 0;
  source->channels = 2;
  
  
  /* By default location is set to LOCAL (stream.c:83) */
  if (source->location == REMOTE) {
    return http_open_location(source);
  } 
  
  /* source->locate == LOCAL */

  if (source->path[0] == '-' && source->path[1] == '\0') {
    stream404_log_debug("Reading audio from stdin ... now unavailable !");
    return -1;
  } else if ((fd = open (source->path, O_RDONLY)) < 0) {
    stream404_util_strerror (errno, buf, sizeof (buf));
    stream404_log_error ("Error opening: %s", buf);

    return -1;
  }

  source->fd = fd;

  if ((rc = lseek (fd, 0, SEEK_END)) >= 0) {
    source->filesize = rc;
    lseek (fd, 0, SEEK_SET);
  }

  if ((len = read (fd, buf, sizeof (buf))) <= 0) {
    stream404_util_strerror (errno, buf, sizeof (buf));
    stream404_log_error ("Error reading header: %s", source->path, buf);

    close (fd);
    return -1;
  }

  /* test extension to accelerate (code comes from xmms) */
  ext = strrchr(source->path, '.'); 
  if (ext) {

#ifdef HAVE_LIBFAAD
    /* peace of code from frederic fondriest */
    if (!strcasecmp(ext, ".mp4") ||	/* official extention */
	!strcasecmp(ext, ".m4a") ||	/* Apple mp4 extention */
	!strcasecmp(ext, ".aac")) {	/* old MPEG2/4-AAC extention */
      if (!(rc = stream404_mp4_open (source, buf, len)))
	return 0;
    }
#endif
    
#ifdef HAVE_LIBFLAC
    if (!strcasecmp(ext, ".flac")) {
      if (!(rc = stream404_flac_open (source, buf, len)))
	return 0;
    }
#endif
	
    if (!strcasecmp(ext, ".mp3")) {
      if (!(rc = stream404_mp3_open (source, buf, len)))
	return 0;
    }
      
    if (!strcasecmp(ext, ".ogg")) {
      if (!(rc = stream404_vorbis_open (source, buf, len)))
	return 0;
    }      

  } else {
    stream404_log_error("no file extension");
  }

  close (fd);
  return -1;
}

/* return how much stream needs to sleep at min before sending data */
/* XXX this function is now obsoleted by timing_sleep()
   static int 
   stream_get_sync(stream404_config_t* config)
   {
   stream404_stream_t* stream;
   int sleeptime = 0;
   int delay;
   for (stream = config->streams; stream; stream = stream->next) {
   delay = shout_delay(stream->conn);
   sleeptime = (sleeptime < delay)? delay : sleeptime;
   }
   return sleeptime;
   }
*/

/* from libshout2 */
void timing_sleep(int sleeptime)
{
  /*
    struct timeval sleeper;
  */
  struct timespec sleeper;
  if (sleeptime <= 0)
    return;

  sleeper.tv_sec = sleeptime / 1000;
  sleeper.tv_nsec = (sleeptime % 1000) * 1000 * 1000;

  /* now try to nanosleep() instead of select() */
  /* sleeptime is in milliseconds */
  nanosleep(&sleeper, NULL);
  
  /*
    sleeper.tv_sec = sleeptime / 1000;
    sleeper.tv_usec = (sleeptime % 1000) * 1000;
    
    * NOTE:
    * This should be 0 for the first argument.  The linux manpage
    * says so.  The solaris manpage also says this is a legal
    * value.  If you think differerntly, please provide references.
    *
    select(0, NULL, NULL, NULL, &sleeper);
  */
}

/* wrapper for shout_send_data, shout_sleep with error handling */
static int
stream_send_data (stream404_stream_t* stream, unsigned char* buf, size_t len)
{
  if (shout_get_connected (stream->conn) != SHOUTERR_CONNECTED) {
    stream_connect (stream);
    if (shout_get_connected (stream->conn) == SHOUTERR_CONNECTED)
      stream404_metadata_update (1);
    else
      return -1;
  }

 
  shout_sync(stream->conn);
  if (shout_send (stream->conn, buf, len) == SHOUTERR_SUCCESS) {
    stream->errs = 0;

    return 0;
  }

  stream404_log_error ("Libshout reported send error, disconnecting: %s",
		  shout_get_error (stream->conn));
  shout_close (stream->conn);
  stream->errs++;

  return -1;
}

static int
stream_connect (stream404_stream_t* stream)
{
  time_t now = time(NULL);
  const char* mount = shout_get_mount (stream->conn);

  if (stream->connect_delay > now)
    return -1;

  if (shout_open (stream->conn) != SHOUTERR_SUCCESS) {
    stream404_log_error ("Mount failed on http://%s:%d%s%s, error: %s",
		    shout_get_host (stream->conn), shout_get_port (stream->conn),
		    (mount && mount[0] == '/') ? "" : "/", stream404_util_nullcheck (mount),
		    shout_get_error (stream->conn));
    stream->connect_delay = now + 1;
    stream->errs++;

    return -1;
  }

  stream404_log ("Mounted on http://%s:%d%s%s", shout_get_host (stream->conn),
            shout_get_port (stream->conn),
            (mount && mount[0] == '/') ? "" : "/", stream404_util_nullcheck (mount));

  return 0;
}

static int stream_needs_reencoding (input_stream_t* source, stream404_stream_t* stream) {
  /*
    if (!source->read || source->bitrate != (unsigned int)stream->bitrate
        || (stream->out_samplerate > 0 &&
	    source->samplerate != (unsigned int)stream->out_samplerate)
        || (stream->out_numchannels > 0 &&
	    source->channels != (unsigned int)stream->out_numchannels))
      return 1;

  return 0;
  */
  return 1;
}
