/* $Id: stream404_config.c,v 1.3 2005/06/26 14:53:44 bedlam Exp $ */
/* parse.c
 * - Functions for xml file parsing
 * Copyright (c) 2000 Alexander Havng
 * Copyright (c) 2001-4 Brendan Cully
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id: stream404_config.c,v 1.3 2005/06/26 14:53:44 bedlam Exp $
 */

#include "definitions.h"

#include <libxml/parser.h>
#include <libxml/xmlmemory.h>

/* Private function declarations */
static int parse_file (const char *configfile, stream404_config_t *stream404_config);
static void parse_playlist_node (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur, stream404_config_t *stream404_config);
static void parse_backup_playlist_node (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur, stream404_config_t *stream404_config);
static void parse_crossfade_node (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur, stream404_config_t *stream404_config);
static void parse_execution_node (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur, stream404_config_t *stream404_config);
static void parse_sql_node (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur, stream404_config_t *stream404_config);
static void parse_server_node (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur,
				stream404_stream_t *stream404_config);
static void parse_stream_node (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur, stream404_stream_t *stream);
static char* stream404_xml_read_node (xmlDocPtr doc, xmlNodePtr node);

/* Global function definitions */

/* Top level XML configfile parser */
int
stream404_xml_parse_config_file (stream404_config_t *stream404_config, const char *configfile)
{
  char namespace[1024];

  if (!stream404_util_verify_file (configfile)) {
    stream404_log_error ("XML Parser Error: Could not open configfile. [%s]  Error: [%s]", configfile, stream404_util_strerror (errno, namespace, 1024));
    return 0;
  }

  /* Setup the parser options */
  LIBXML_TEST_VERSION
  xmlKeepBlanksDefault(0);
	
  /* Parse the file and be happy */
  return parse_file (configfile, stream404_config);
}

/* I hope you can tell this is my first try at xml and libxml :)  */
static int
parse_file (const char *configfile, stream404_config_t *stream404_config)
{
  xmlDocPtr doc = NULL;
  xmlNsPtr ns = NULL;
  xmlNodePtr cur = NULL;

  /* we support multiple stream nodes */
  int nstreams = 0;
  stream404_stream_t* stream = stream404_config->streams;

  /* This does the actual parsing */
  if (!(doc = xmlParseFile (configfile))) {
    stream404_log_error ("XML Parser: Error while parsing %s", configfile);
    return 0;
  }
	
  /* Gimme the root element dammit! */
  if (!(cur = xmlDocGetRootElement (doc))) {
    stream404_log_error ("XML Parser: Empty documenent %s", configfile);
    xmlFreeDoc (doc);
    return 0;
  }
	
  /* Verify that the document is of the right type in the right namespace 
  if (!(ns = xmlSearchNsByHref (doc, cur, (unsigned char *)"http://www.icecast.org/projects/stream404"))) {
    stream404_log_error ("XML Parser: Document of invalid type, no stream404 namespace found");
    xmlFreeDoc (doc);
    return 0;
  }
  */

  /* First element should be configuration */
  if (!cur->name || (strcmp (cur->name, "Configuration"))) {
    stream404_log_error ("XML Parser: Document of invalid type, root node is not 'Configuration'");
    xmlFreeDoc (doc);
    return 0;
  }

  /* Get the configuration tree */
  /* Tree traversal */
  cur = cur->xmlChildrenNode;

#ifdef HAVE_LIBXML_PARSER_H
  while (cur && xmlIsBlankNode (cur)) {
    cur = cur->next;
  }
#endif

  if (cur == NULL) {
    /* Right type of document, but no configuration, guess that is ok */
    xmlFreeDoc (doc);
    return 1;
  }

  /* Separate the parsing into different submodules,
   * Server, Stream, Playlist and Execution */
  for (; cur != NULL; cur = cur->next) {
    if (cur->type == XML_COMMENT_NODE)
      continue;
    if (strcmp (cur->name, "Stream") == 0) {
      /* first stream is preallocated */
      if (nstreams) {
	stream->next = (stream404_stream_t*)malloc (sizeof (stream404_stream_t));
	stream = stream->next;
	/* in case fields are omitted in the config file */
	stream404_setup_parse_stream_defaults (stream);
      }

      parse_stream_node (doc, ns, cur->xmlChildrenNode, stream);

      nstreams++;
    } else if (strcmp (cur->name, "Playlist") == 0) {
      parse_playlist_node (doc, ns, cur->xmlChildrenNode, stream404_config);
    } else if (strcmp (cur->name, "BackupPlaylist") == 0) {
      parse_backup_playlist_node (doc, ns, cur->xmlChildrenNode, stream404_config);
    } else if (strcmp (cur->name, "Execution") == 0) {
      parse_execution_node (doc, ns, cur->xmlChildrenNode, stream404_config);
    } else if (strcmp (cur->name, "Sql") == 0) {
      parse_sql_node (doc, ns, cur->xmlChildrenNode, stream404_config);
    } else {
      stream404_log ("Unknown Node: %s", cur->name);
    }
  }
	
  /* Be a good boy and cleanup */
  xmlFreeDoc (doc);
  return 1;
}

/* Parse the stream specific configuration */
static void
parse_stream_node (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur,
			     stream404_stream_t *stream)
{
  for (; cur; cur = cur->next) {
    if (cur->type == XML_COMMENT_NODE)
      continue;

    if (strcmp (cur->name, "Server") == 0) {
      parse_server_node (doc, ns, cur->xmlChildrenNode, stream);
    } else if (strcmp (cur->name, "Name") == 0) {
      stream404_util_free (stream->name);
      stream->name = stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Genre") == 0) {
      stream404_util_free (stream->genre);
      stream->genre = stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Description") == 0) {
      stream404_util_free (stream->description);
      stream->description = stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "URL") == 0) {
      stream404_util_free (stream->url);
      stream->url = stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Mountpoint") == 0) {
      stream404_util_free (stream->mount);
      stream->mount = stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Dumpfile") == 0) {
      stream404_util_free (stream->dumpfile);
      stream->dumpfile = stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Bitrate") == 0) {
      stream->bitrate = atoi (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Quality") == 0) {
      stream->quality = atof (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Public") == 0) {
      stream->ispublic = atoi (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Reencode") == 0) {
      int res = atoi (stream404_xml_read_node (doc, cur));

#ifndef HAVE_LIBLAME
      if (res == 1) {
	stream404_log ("Support for reencoding with liblame was not found. You can't reencode this.");
	stream404_setup_shutdown ();
      }
#endif
			
      stream->reencode = res;
    } else if (strcmp (cur->name, "Samplerate") == 0) {
      stream->out_samplerate = atoi (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Channels") == 0) {
      stream->out_numchannels = atoi (stream404_xml_read_node (doc, cur));
    } else {
      stream404_log ("Unknown Stream keyword: %s", cur->name);
    }
  }

   /* check for arguments */
  if ((stream->type == STREAM404_OUTPUT_VORBIS) && 
	(stream->quality == STREAM404_DEFAULT_QUALITY)) {
    stream404_log ("When streaming with ogg/vorbis you have to setup quality instead of bitrate");
    stream404_setup_shutdown ();
  }
}

/* Parse the server specific configuration */
static void
parse_server_node (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur,
		   stream404_stream_t *stream404_config)
{
  for (; cur; cur = cur->next) {
    if (cur->type == XML_COMMENT_NODE)
      continue;

    if (strcmp (cur->name, "Port") == 0) {
      stream404_config->port = atoi (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Hostname") == 0) {
      stream404_util_free (stream404_config->host);
      stream404_config->host = stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Password") == 0) {
      stream404_util_free (stream404_config->password);
      stream404_config->password = stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Protocol") == 0) {
      unsigned char *str = stream404_xml_read_node (doc, cur);

      if (str && (strcasecmp(str, "icy") == 0))
	stream404_config->protocol = icy_protocol_e;
      else if (str && (strcasecmp(str, "http") == 0))
	stream404_config->protocol = http_protocol_e;
      else
	stream404_config->protocol = xaudiocast_protocol_e;
    } else if (strcmp (cur->name, "Format") == 0) {
      unsigned char *str = stream404_xml_read_node (doc, cur);

      if (str && (strcasecmp(str, "ogg") == 0))
	stream404_config->type = STREAM404_OUTPUT_VORBIS;
      else 
	stream404_config->type = STREAM404_OUTPUT_MP3;
    } else {
      stream404_log ("Unknown Server keyword: %s", cur->name);
    }
  }

  /* check for arguments */
  if ((stream404_config->type == STREAM404_OUTPUT_VORBIS) && 
      (stream404_config->protocol == icy_protocol_e)) {
    stream404_log ("Cannot stream Vorbis to shoutcast server");
    stream404_setup_shutdown ();
  }
}

/* Parse the execution specific options */
static void
parse_execution_node (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur,
				stream404_config_t *stream404_config)
{
  for (; cur; cur = cur->next) {
    if (cur->type == XML_COMMENT_NODE)
      continue;

    if (strcmp (cur->name, "Background") == 0) {
      stream404_config->daemon = atoi (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Verbose") == 0) {
      stream404_config->verbose = atoi (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "BaseDirectory") == 0) {
      if (stream404_config->base_directory)
	stream404_config->base_directory =
	  stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else {
      stream404_log ("Unknown Execution keyword: %s", cur->name);
    }
  }
}

/* Parse the sql specific options */
static void
parse_sql_node (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur,
		      stream404_config_t *stream404_config)
{
  for (; cur; cur = cur->next) {
    if (cur->type == XML_COMMENT_NODE)
      continue;

    if (strcmp (cur->name, "Hostname") == 0) {
      stream404_util_free (stream404_config->sql.hostname);
      stream404_config->sql.hostname = 
	stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Port") == 0) {
      stream404_config->sql.port = atoi (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Unixsocket") == 0) {
      stream404_util_free (stream404_config->sql.usocket);
      stream404_config->sql.usocket = 
	stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Sqltype") == 0) {
      stream404_util_free (stream404_config->sql.sqltype);
      stream404_config->sql.sqltype =  
	stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "User") == 0) {
      stream404_util_free (stream404_config->sql.user);
      stream404_config->sql.user = stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Password") == 0) {
      stream404_util_free (stream404_config->sql.password);
      stream404_config->sql.password = 
	stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Database") == 0) {
      stream404_util_free (stream404_config->sql.database);
      stream404_config->sql.database = 
	stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Table") == 0) {
      stream404_util_free (stream404_config->sql.table);
      stream404_config->sql.table =
	stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "UrgTable") == 0) {
      stream404_util_free (stream404_config->sql.urg_table);
      stream404_config->sql.urg_table =
	stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "CurrentStreamTable") == 0) {
      stream404_util_free (stream404_config->sql.currentstream_table);
      stream404_config->sql.currentstream_table =
	stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "StreamLogTable") == 0) {
      stream404_util_free (stream404_config->sql.streamlog_table);
      stream404_config->sql.streamlog_table =
	stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "MetadataTable") == 0) {
      stream404_util_free (stream404_config->sql.metadata_table);
      stream404_config->sql.metadata_table =
	stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else {
      stream404_log ("Unknown Execution keyword: %s", cur->name);
    }
  }
}

/* Parse the playlist specific configuration */
static void
parse_playlist_node (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur, stream404_config_t *stream404_config)
{
  for (; cur; cur = cur->next) {
    if (cur->type == XML_COMMENT_NODE)
      continue;

    if (strcmp (cur->name, "Randomize") == 0) {
      stream404_config->pm.randomize = atoi (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Type") == 0) {
      unsigned char *str = stream404_xml_read_node (doc, cur);
      if (str && (strcmp (str, "python") == 0))
	stream404_config->pm.playlist_type = stream404_playlist_python_e;
      else if (str && (strcmp (str, "perl") == 0))
	stream404_config->pm.playlist_type = stream404_playlist_perl_e;
      else if (str && (strcmp (str, "builtin") == 0))
	stream404_config->pm.playlist_type = stream404_playlist_builtin_e;
      else
	stream404_config->pm.playlist_type = stream404_playlist_sql_e;
    } else if (strcmp (cur->name, "File") == 0) {
      stream404_util_free (stream404_config->pm.playlist_file);
      stream404_config->pm.playlist_file =
	stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Module") == 0) {
      stream404_util_free (stream404_config->pm.module);
      stream404_config->pm.module =
	stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Crossfade") == 0) {
	parse_crossfade_node (doc, ns, cur->xmlChildrenNode, stream404_config);
    } else {
      stream404_log ("Unknown playlist keyword: %s", cur->name);
    }
  }		
}

/* Parse the backup playlist specific configuration */
static void
parse_backup_playlist_node (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur, stream404_config_t *stream404_config)
{
  for (; cur; cur = cur->next) {
    if (cur->type == XML_COMMENT_NODE)
      continue;

    if (strcmp (cur->name, "Randomize") == 0) {
      stream404_config->pm_backup.randomize = atoi (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Type") == 0) {
      unsigned char *str = stream404_xml_read_node (doc, cur);
      if (str && (strcmp (str, "python") == 0))
	stream404_config->pm_backup.playlist_type = stream404_playlist_python_e;
      else if (str && (strcmp (str, "perl") == 0))
	stream404_config->pm_backup.playlist_type = stream404_playlist_perl_e;
      else if (str && (strcmp (str, "builtin") == 0))
	stream404_config->pm_backup.playlist_type = stream404_playlist_builtin_e;
      else
	stream404_config->pm_backup.playlist_type = stream404_playlist_sql_e;
    } else if (strcmp (cur->name, "File") == 0) {
      stream404_util_free (stream404_config->pm_backup.playlist_file);
      stream404_config->pm_backup.playlist_file =
	stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Module") == 0) {
      stream404_util_free (stream404_config->pm_backup.module);
      stream404_config->pm_backup.module =
	stream404_util_strdup (stream404_xml_read_node (doc, cur));
    } else if (strcmp (cur->name, "Crossfade") == 0) {
	parse_crossfade_node (doc, ns, cur->xmlChildrenNode, stream404_config);
    } else {
      stream404_log ("Unknown backup playlist keyword: %s", cur->name);
    }
  }		
}

static void
parse_crossfade_node (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur, stream404_config_t *stream404_config)
{
    int fadelen, fileminsecs;
    fadelen = fileminsecs = 0;
    
    for (; cur; cur = cur->next) {
	if (cur->type == XML_COMMENT_NODE)
	    continue;
	
	if (strcmp (cur->name, "Fadelen") == 0) {
	    fadelen = atoi (stream404_xml_read_node (doc, cur));
	} else if (strcmp (cur->name, "MinSecs") == 0) {
	    fileminsecs = atoi (stream404_xml_read_node (doc, cur));
	}
    }
    
    if (fadelen > 0) {
	if (fileminsecs > 0) {
	    stream404_config->plugins = 
		crossfade_plugin(fadelen, fileminsecs);
	} else {
	    stream404_config->plugins = 
		 crossfade_plugin(fadelen, 0);
	}
    }
    
}

static char* stream404_xml_read_node (xmlDocPtr doc, xmlNodePtr node)
{
  return (char *)xmlNodeListGetString (doc, node->xmlChildrenNode, 1);
}
